import jax.numpy as jnp
import jaxdecomp as jd

# Pure functions for hydro simulation


def get_conserved(rho, vx, vy, vz, vol):
    Mass = rho * vol
    Momx = rho * vx * vol
    Momy = rho * vy * vol
    Momz = rho * vz * vol

    return Mass, Momx, Momy, Momz


def get_primitive(Mass, Momx, Momy, Momz, vol):
    rho = Mass / vol
    vx = Momx / Mass
    vy = Momy / Mass
    vz = Momz / Mass

    return rho, vx, vy, vz


def get_gradient(f, dx):
    f_dx = (jnp.roll(f, -1, axis=0) - jnp.roll(f, 1, axis=0)) / (2.0 * dx)
    f_dy = (jnp.roll(f, -1, axis=1) - jnp.roll(f, 1, axis=1)) / (2.0 * dx)
    f_dz = (jnp.roll(f, -1, axis=2) - jnp.roll(f, 1, axis=2)) / (2.0 * dx)

    return f_dx, f_dy, f_dz


def extrap_to_face(f, f_dx, f_dy, f_dz, dx):
    f_XL = f + 0.5 * f_dx * dx
    f_XR = f - 0.5 * f_dx * dx
    f_XR = jnp.roll(f_XR, -1, axis=0)

    f_YL = f + 0.5 * f_dy * dx
    f_YR = f - 0.5 * f_dy * dx
    f_YR = jnp.roll(f_YR, -1, axis=1)

    f_ZL = f + 0.5 * f_dz * dx
    f_ZR = f - 0.5 * f_dz * dx
    f_ZR = jnp.roll(f_ZR, -1, axis=2)

    return f_XL, f_XR, f_YL, f_YR, f_ZL, f_ZR


def apply_fluxes(F, flux_F_X, flux_F_Y, flux_F_Z, dx, dt):
    fac = dt * dx * dx
    F += -fac * flux_F_X
    F += fac * jnp.roll(flux_F_X, 1, axis=0)

    F += -fac * flux_F_Y
    F += fac * jnp.roll(flux_F_Y, 1, axis=1)

    F += -fac * flux_F_Z
    F += fac * jnp.roll(flux_F_Z, 1, axis=2)

    return F


def get_flux(rho_L, vx_L, vy_L, vz_L, rho_R, vx_R, vy_R, vz_R, cs):
    # compute star (averaged) states
    rho_star = 0.5 * (rho_L + rho_R)
    momx_star = 0.5 * (rho_L * vx_L + rho_R * vx_R)
    momy_star = 0.5 * (rho_L * vy_L + rho_R * vy_R)
    momz_star = 0.5 * (rho_L * vz_L + rho_R * vz_R)

    P_star = rho_star * cs * cs

    # compute fluxes (local Lax-Friedrichs/Rusanov)
    flux_Mass = momx_star
    flux_Momx = momx_star**2 / rho_star + P_star
    flux_Momy = momx_star * momy_star / rho_star
    flux_Momz = momx_star * momz_star / rho_star

    # find wavespeeds
    C_L = cs + jnp.abs(vx_L)
    C_R = cs + jnp.abs(vx_R)
    C = jnp.maximum(C_L, C_R)

    # add stabilizing diffusive term
    flux_Mass -= C * 0.5 * (rho_R - rho_L)
    flux_Momx -= C * 0.5 * (rho_R * vx_R - rho_L * vx_L)
    flux_Momy -= C * 0.5 * (rho_R * vy_R - rho_L * vy_L)
    flux_Momz -= C * 0.5 * (rho_R * vz_R - rho_L * vz_L)

    return flux_Mass, flux_Momx, flux_Momy, flux_Momz


def hydro_accelerate(vx, vy, vz, V, kx, ky, kz, dt):
    V_hat = jd.fft.pfft3d(V)

    ax = -jnp.real(jd.fft.pifft3d(1.0j * kx * V_hat))
    ay = -jnp.real(jd.fft.pifft3d(1.0j * ky * V_hat))
    az = -jnp.real(jd.fft.pifft3d(1.0j * kz * V_hat))

    vx += ax * dt
    vy += ay * dt
    vz += az * dt

    return vx, vy, vz


def hydro_fluxes(rho, vx, vy, vz, dt, dx, cs):
    # calculate gradients
    rho_dx, rho_dy, rho_dz = get_gradient(rho, dx)
    vx_dx, vx_dy, vx_dz = get_gradient(vx, dx)
    vy_dx, vy_dy, vy_dz = get_gradient(vy, dx)
    vz_dx, vz_dy, vz_dz = get_gradient(vz, dx)

    # slope limit gradients
    rho_dx, rho_dy, rho_dz = slope_limiter(rho, dx, rho_dx, rho_dy, rho_dz)
    vx_dx, vx_dy, vx_dz = slope_limiter(vx, dx, vx_dx, vx_dy, vx_dz)
    vy_dx, vy_dy, vy_dz = slope_limiter(vy, dx, vy_dx, vy_dy, vy_dz)
    vz_dx, vz_dy, vz_dz = slope_limiter(vz, dx, vz_dx, vz_dy, vz_dz)

    # extrapolate half-step in time
    rho_prime = rho - 0.5 * dt * (
        vx * rho_dx
        + rho * vx_dx
        + vy * rho_dy
        + rho * vy_dy
        + vz * rho_dz
        + rho * vz_dz
    )
    vx_prime = vx - 0.5 * dt * (
        vx * vx_dx + vy * vx_dy + vz * vx_dz + (1.0 / rho) * rho_dx * cs * cs
    )
    vy_prime = vy - 0.5 * dt * (
        vx * vy_dx + vy * vy_dy + vz * vy_dz + (1.0 / rho) * rho_dy * cs * cs
    )
    vz_prime = vz - 0.5 * dt * (
        vx * vz_dx + vy * vz_dy + vz * vz_dz + (1.0 / rho) * rho_dz * cs * cs
    )

    # extrapolate in space to face centers
    rho_XL, rho_XR, rho_YL, rho_YR, rho_ZL, rho_ZR = extrap_to_face(
        rho_prime, rho_dx, rho_dy, rho_dz, dx
    )
    vx_XL, vx_XR, vx_YL, vx_YR, vx_ZL, vx_ZR = extrap_to_face(
        vx_prime, vx_dx, vx_dy, vx_dz, dx
    )
    vy_XL, vy_XR, vy_YL, vy_YR, vy_ZL, vy_ZR = extrap_to_face(
        vy_prime, vy_dx, vy_dy, vy_dz, dx
    )
    vz_XL, vz_XR, vz_YL, vz_YR, vz_ZL, vz_ZR = extrap_to_face(
        vz_prime, vz_dx, vz_dy, vz_dz, dx
    )

    # compute fluxes (local Lax-Friedrichs/Rusanov)
    flux_Mass_X, flux_Momx_X, flux_Momy_X, flux_Momz_X = get_flux(
        rho_XL, vx_XL, vy_XL, vz_XL, rho_XR, vx_XR, vy_XR, vz_XR, cs
    )
    flux_Mass_Y, flux_Momy_Y, flux_Momz_Y, flux_Momx_Y = get_flux(
        rho_YL, vy_YL, vz_YL, vx_YL, rho_YR, vy_YR, vz_YR, vx_YR, cs
    )
    flux_Mass_Z, flux_Momz_Z, flux_Momx_Z, flux_Momy_Z = get_flux(
        rho_ZL, vz_ZL, vx_ZL, vy_ZL, rho_ZR, vz_ZR, vx_ZR, vy_ZR, cs
    )

    Mass, Momx, Momy, Momz = get_conserved(rho, vx, vy, vz, dx**3)

    # update solution
    Mass = apply_fluxes(Mass, flux_Mass_X, flux_Mass_Y, flux_Mass_Z, dx, dt)
    Momx = apply_fluxes(Momx, flux_Momx_X, flux_Momx_Y, flux_Momx_Z, dx, dt)
    Momy = apply_fluxes(Momy, flux_Momy_X, flux_Momy_Y, flux_Momy_Z, dx, dt)
    Momz = apply_fluxes(Momz, flux_Momz_X, flux_Momz_Y, flux_Momz_Z, dx, dt)

    # get Primitive variables
    rho, vx, vy, vz = get_primitive(Mass, Momx, Momy, Momz, dx**3)

    return rho, vx, vy, vz


def slope_limiter(f, dx, f_dx, f_dy, f_dz):
    """
    Apply slope limiter to slopes (minmod)
    """

    eps = 1.0e-12

    # Keep a copy of the original slopes
    orig_f_dx = f_dx
    orig_f_dy = f_dy
    orig_f_dz = f_dz

    # Function to adjust the denominator safely
    def adjust_denominator(denom):
        denom_safe = jnp.where(
            denom > 0, denom + eps, jnp.where(denom < 0, denom - eps, eps)
        )
        return denom_safe

    # For x-direction
    denom = adjust_denominator(orig_f_dx)
    num = (f - jnp.roll(f, 1, axis=0)) / dx
    ratio = num / denom
    limiter = jnp.maximum(0.0, jnp.minimum(1.0, ratio))
    f_dx = limiter * f_dx

    num = -(f - jnp.roll(f, -1, axis=0)) / dx
    ratio = num / denom  # Use the same adjusted denominator
    limiter = jnp.maximum(0.0, jnp.minimum(1.0, ratio))
    f_dx = limiter * f_dx

    # For y-direction
    denom = adjust_denominator(orig_f_dy)
    num = (f - jnp.roll(f, 1, axis=1)) / dx
    ratio = num / denom
    limiter = jnp.maximum(0.0, jnp.minimum(1.0, ratio))
    f_dy = limiter * f_dy

    num = -(f - jnp.roll(f, -1, axis=1)) / dx
    ratio = num / denom
    limiter = jnp.maximum(0.0, jnp.minimum(1.0, ratio))
    f_dy = limiter * f_dy

    # For z-direction
    denom = adjust_denominator(orig_f_dz)
    num = (f - jnp.roll(f, 1, axis=2)) / dx
    ratio = num / denom
    limiter = jnp.maximum(0.0, jnp.minimum(1.0, ratio))
    f_dz = limiter * f_dz

    num = -(f - jnp.roll(f, -1, axis=2)) / dx
    ratio = num / denom
    limiter = jnp.maximum(0.0, jnp.minimum(1.0, ratio))
    f_dz = limiter * f_dz

    return f_dx, f_dy, f_dz



import jax.numpy as jnp
import jaxdecomp as jd

# Pure functions for quantum simulation


def quantum_kick(psi, V, m_per_hbar, dt):
    psi = jnp.exp(-1.0j * m_per_hbar * dt * V) * psi
    return psi


def quantum_drift(psi, k_sq, m_per_hbar, dt):
    psi_hat = jd.fft.pfft3d(psi)
    psi_hat = jnp.exp(dt * (-1.0j * k_sq / m_per_hbar / 2.0)) * psi_hat
    psi = jd.fft.pifft3d(psi_hat)
    return psi


def get_gradient(psi, kx, ky, kz):
    """
    Computes gradient of wavefunction psi
    psi: jnp.ndarray (3D)
    Returns: grad_psi_x, grad_psi_y, grad_psi_z
    """
    psi_hat = jd.fft.pfft3d(psi)
    grad_psi_x_hat = 1.0j * kx * psi_hat
    grad_psi_y_hat = 1.0j * ky * psi_hat
    grad_psi_z_hat = 1.0j * kz * psi_hat

    grad_psi_x = jd.fft.pifft3d(grad_psi_x_hat)
    grad_psi_y = jd.fft.pifft3d(grad_psi_y_hat)
    grad_psi_z = jd.fft.pifft3d(grad_psi_z_hat)

    return grad_psi_x, grad_psi_y, grad_psi_z


def quantum_velocity(psi, box_size, m_per_hbar):
    """
    Compute the velocity from the wave-function
    v = nabla S / m
    psi = sqrt(rho) exp(i S / hbar)
    """
    N = psi.shape[0]
    dx = box_size / N

    S_per_hbar = jnp.angle(psi)

    # Central differences with phase unwrapping
    # vx = jnp.roll(S_per_hbar, -1, axis=0) - jnp.roll(S_per_hbar, 1, axis=0)
    # vy = jnp.roll(S_per_hbar, -1, axis=1) - jnp.roll(S_per_hbar, 1, axis=1)
    # vz = jnp.roll(S_per_hbar, -1, axis=2) - jnp.roll(S_per_hbar, 1, axis=2)
    # vx = jnp.where(vx > jnp.pi, vx - 2 * jnp.pi, vx)
    # vx = jnp.where(vx <= -jnp.pi, vx + 2 * jnp.pi, vx)
    # vy = jnp.where(vy > jnp.pi, vy - 2 * jnp.pi, vy)
    # vy = jnp.where(vy <= -jnp.pi, vy + 2 * jnp.pi, vy)
    # vz = jnp.where(vz > jnp.pi, vz - 2 * jnp.pi, vz)
    # vz = jnp.where(vz <= -jnp.pi, vz + 2 * jnp.pi, vz)
    # vx = vx / (2.0 * dx) / m_per_hbar
    # vy = vy / (2.0 * dx) / m_per_hbar
    # vz = vz / (2.0 * dx) / m_per_hbar

    # Forward differences
    vx1 = jnp.roll(S_per_hbar, -1, axis=0) - S_per_hbar
    vy1 = jnp.roll(S_per_hbar, -1, axis=1) - S_per_hbar
    vz1 = jnp.roll(S_per_hbar, -1, axis=2) - S_per_hbar
    vx1 = jnp.where(vx1 > jnp.pi, vx1 - 2 * jnp.pi, vx1)
    vx1 = jnp.where(vx1 <= -jnp.pi, vx1 + 2 * jnp.pi, vx1)
    vy1 = jnp.where(vy1 > jnp.pi, vy1 - 2 * jnp.pi, vy1)
    vy1 = jnp.where(vy1 <= -jnp.pi, vy1 + 2 * jnp.pi, vy1)
    vz1 = jnp.where(vz1 > jnp.pi, vz1 - 2 * jnp.pi, vz1)
    vz1 = jnp.where(vz1 <= -jnp.pi, vz1 + 2 * jnp.pi, vz1)
    vx1 = vx1 / dx / m_per_hbar
    vy1 = vy1 / dx / m_per_hbar
    vz1 = vz1 / dx / m_per_hbar

    # Backward differences
    vx2 = S_per_hbar - jnp.roll(S_per_hbar, 1, axis=0)
    vy2 = S_per_hbar - jnp.roll(S_per_hbar, 1, axis=1)
    vz2 = S_per_hbar - jnp.roll(S_per_hbar, 1, axis=2)
    vx2 = jnp.where(vx2 > jnp.pi, vx2 - 2 * jnp.pi, vx2)
    vx2 = jnp.where(vx2 <= -jnp.pi, vx2 + 2 * jnp.pi, vx2)
    vy2 = jnp.where(vy2 > jnp.pi, vy2 - 2 * jnp.pi, vy2)
    vy2 = jnp.where(vy2 <= -jnp.pi, vy2 + 2 * jnp.pi, vy2)
    vz2 = jnp.where(vz2 > jnp.pi, vz2 - 2 * jnp.pi, vz2)
    vz2 = jnp.where(vz2 <= -jnp.pi, vz2 + 2 * jnp.pi, vz2)
    vx2 = vx2 / dx / m_per_hbar
    vy2 = vy2 / dx / m_per_hbar
    vz2 = vz2 / dx / m_per_hbar

    # Average forward and backward
    vx = 0.5 * (vx1 + vx2)
    vy = 0.5 * (vy1 + vy2)
    vz = 0.5 * (vz1 + vz2)

    return vx, vy, vz



import jax
import jax.numpy as jnp
import jaxdecomp as jd

# Pure functions for particle-mesh calculations (stars, BHs, ...)


def get_cic_indices_and_weights(pos, dx, resolution):
    """Compute the cloud-in-cell indices and weights for the particle positions."""
    nx = resolution
    dxs = jnp.array([dx, dx, dx])
    i = jnp.floor((pos - 0.5 * dxs) / dxs)
    ip1 = i + 1.0
    weight_i = ((ip1 + 0.5) * dxs - pos) / dxs
    weight_ip1 = (pos - (i + 0.5) * dxs) / dxs
    i = jnp.mod(i, jnp.array([nx, nx, nx])).astype(int)
    ip1 = jnp.mod(ip1, jnp.array([nx, nx, nx])).astype(int)
    return i, ip1, weight_i, weight_ip1


def bin_particles(pos, m_particles, dx, resolution, multiple_masses):
    """Bin the particles into the grid using cloud-in-cell weights."""
    nx = resolution
    n_particle = pos.shape[0]
    rho = jnp.zeros((nx, nx, nx))
    i, ip1, w_i, w_ip1 = get_cic_indices_and_weights(pos, dx, resolution)

    def deposit_particle(s, rho):
        """Deposit the particle mass into the grid."""
        if multiple_masses:
            m_particle = m_particles[s]
        else:
            m_particle = m_particles
        fac = m_particle / (dx * dx * dx)
        rho = rho.at[i[s, 0], i[s, 1], i[s, 2]].add(
            w_i[s, 0] * w_i[s, 1] * w_i[s, 2] * fac
        )
        rho = rho.at[ip1[s, 0], i[s, 1], i[s, 2]].add(
            w_ip1[s, 0] * w_i[s, 1] * w_i[s, 2] * fac
        )
        rho = rho.at[i[s, 0], ip1[s, 1], i[s, 2]].add(
            w_i[s, 0] * w_ip1[s, 1] * w_i[s, 2] * fac
        )
        rho = rho.at[i[s, 0], i[s, 1], ip1[s, 2]].add(
            w_i[s, 0] * w_i[s, 1] * w_ip1[s, 2] * fac
        )
        rho = rho.at[ip1[s, 0], ip1[s, 1], i[s, 2]].add(
            w_ip1[s, 0] * w_ip1[s, 1] * w_i[s, 2] * fac
        )
        rho = rho.at[ip1[s, 0], i[s, 1], ip1[s, 2]].add(
            w_ip1[s, 0] * w_i[s, 1] * w_ip1[s, 2] * fac
        )
        rho = rho.at[i[s, 0], ip1[s, 1], ip1[s, 2]].add(
            w_i[s, 0] * w_ip1[s, 1] * w_ip1[s, 2] * fac
        )
        rho = rho.at[ip1[s, 0], ip1[s, 1], ip1[s, 2]].add(
            w_ip1[s, 0] * w_ip1[s, 1] * w_ip1[s, 2] * fac
        )
        return rho

    rho = jax.lax.fori_loop(0, n_particle, deposit_particle, rho)
    return rho


def get_acceleration(pos, V, kx, ky, kz, dx):
    """Compute the acceleration of the particles."""
    n_particle = pos.shape[0]
    resolution = V.shape[0]
    i, ip1, w_i, w_ip1 = get_cic_indices_and_weights(pos, dx, resolution)

    # find accelerations on the grid
    V_hat = jd.fft.pfft3d(V)
    ax = -jnp.real(jd.fft.pifft3d(1.0j * kx * V_hat))
    ay = -jnp.real(jd.fft.pifft3d(1.0j * ky * V_hat))
    az = -jnp.real(jd.fft.pifft3d(1.0j * kz * V_hat))
    a_grid = jnp.stack((ax, ay, az), axis=-1)

    # interpolate the accelerations to the particle positions
    acc = jnp.zeros((n_particle, 3))
    acc += (w_i[:, 0] * w_i[:, 1] * w_i[:, 2])[:, None] * a_grid[
        i[:, 0], i[:, 1], i[:, 2]
    ]
    acc += (w_ip1[:, 0] * w_i[:, 1] * w_i[:, 2])[:, None] * a_grid[
        ip1[:, 0], i[:, 1], i[:, 2]
    ]
    acc += (w_i[:, 0] * w_ip1[:, 1] * w_i[:, 2])[:, None] * a_grid[
        i[:, 0], ip1[:, 1], i[:, 2]
    ]
    acc += (w_i[:, 0] * w_i[:, 1] * w_ip1[:, 2])[:, None] * a_grid[
        i[:, 0], i[:, 1], ip1[:, 2]
    ]
    acc += (w_ip1[:, 0] * w_ip1[:, 1] * w_i[:, 2])[:, None] * a_grid[
        ip1[:, 0], ip1[:, 1], i[:, 2]
    ]
    acc += (w_ip1[:, 0] * w_i[:, 1] * w_ip1[:, 2])[:, None] * a_grid[
        ip1[:, 0], i[:, 1], ip1[:, 2]
    ]
    acc += (w_i[:, 0] * w_ip1[:, 1] * w_ip1[:, 2])[:, None] * a_grid[
        i[:, 0], ip1[:, 1], ip1[:, 2]
    ]
    acc += (w_ip1[:, 0] * w_ip1[:, 1] * w_ip1[:, 2])[:, None] * a_grid[
        ip1[:, 0], ip1[:, 1], ip1[:, 2]
    ]
    return acc


def particles_accelerate(vel, pos, V, kx, ky, kz, dx, dt):
    """Apply the acceleration to the particles."""
    acc = get_acceleration(pos, V, kx, ky, kz, dx)
    vel += acc * dt
    return vel


def particles_drift(pos, vel, dt, box_size):
    pos += vel * dt
    pos = jnp.mod(pos, jnp.array([box_size, box_size, box_size]))

    return pos


def particles_accrete_gas(mass, rho, pos, G, sound_speed, dx, dt):
    """Accrete gas onto particles (Bondi)."""
    n_particle = pos.shape[0]
    resolution = rho.shape[0]
    i, ip1, w_i, w_ip1 = get_cic_indices_and_weights(pos, dx, resolution)
    d_mass = jnp.zeros_like(mass)
    d_rho = jnp.zeros_like(rho)
    lam = jnp.exp(1.5) / 4.0  # ≈ 1.12
    vol = dx**3

    def accrete(s, deltas):
        """Deposit the particle mass into the grid."""
        d_mass, d_rho = deltas
        dM_fac = dt * 4.0 * jnp.pi * lam * (G * mass[s]) ** 2 / sound_speed**3
        # dM = dM_fac * rho

        dm = w_i[s, 0] * w_i[s, 1] * w_i[s, 2] * dM_fac * rho[i[s, 0], i[s, 1], i[s, 2]]
        d_rho = d_rho.at[i[s, 0], i[s, 1], i[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        dm = (
            w_ip1[s, 0]
            * w_i[s, 1]
            * w_i[s, 2]
            * dM_fac
            * rho[ip1[s, 0], i[s, 1], i[s, 2]]
        )
        d_rho = d_rho.at[ip1[s, 0], i[s, 1], i[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        dm = (
            w_i[s, 0]
            * w_ip1[s, 1]
            * w_i[s, 2]
            * dM_fac
            * rho[i[s, 0], ip1[s, 1], i[s, 2]]
        )
        d_rho = d_rho.at[i[s, 0], ip1[s, 1], i[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        dm = (
            w_i[s, 0]
            * w_i[s, 1]
            * w_ip1[s, 2]
            * dM_fac
            * rho[i[s, 0], i[s, 1], ip1[s, 2]]
        )
        d_rho = d_rho.at[i[s, 0], i[s, 1], ip1[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        dm = (
            w_ip1[s, 0]
            * w_ip1[s, 1]
            * w_i[s, 2]
            * dM_fac
            * rho[ip1[s, 0], ip1[s, 1], i[s, 2]]
        )
        d_rho = d_rho.at[ip1[s, 0], ip1[s, 1], i[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        dm = (
            w_ip1[s, 0]
            * w_i[s, 1]
            * w_ip1[s, 2]
            * dM_fac
            * rho[ip1[s, 0], i[s, 1], ip1[s, 2]]
        )
        d_rho = d_rho.at[ip1[s, 0], i[s, 1], ip1[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        dm = (
            w_i[s, 0]
            * w_ip1[s, 1]
            * w_ip1[s, 2]
            * dM_fac
            * rho[i[s, 0], ip1[s, 1], ip1[s, 2]]
        )
        d_rho = d_rho.at[i[s, 0], ip1[s, 1], ip1[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        dm = (
            w_ip1[s, 0]
            * w_ip1[s, 1]
            * w_ip1[s, 2]
            * dM_fac
            * rho[ip1[s, 0], ip1[s, 1], ip1[s, 2]]
        )
        d_rho = d_rho.at[ip1[s, 0], ip1[s, 1], ip1[s, 2]].add(-dm / vol)
        d_mass = d_mass.at[s].add(dm)

        return d_mass, d_rho

    d_mass, d_rho = jax.lax.fori_loop(0, n_particle, accrete, (d_mass, d_rho))
    mass = mass + d_mass
    rho = rho + d_rho

    return mass, rho



import jax.numpy as jnp
from jaxion.utils import run_example_main
import pytest

rel_tol = 1e-4


def test_tidal_stripping():
    sim = run_example_main(
        "examples/tidal_stripping/tidal_stripping.py",
        argv=["--res", "1", "--save", "False"],
    )
    assert sim.resolution == 32
    assert sim.state["t"] > 0.0
    assert jnp.mean(jnp.abs(sim.state["psi"])) == pytest.approx(162.028, rel=rel_tol)


def test_tidal_stripping_distributed_emulate():
    sim = run_example_main(
        "examples/tidal_stripping/tidal_stripping.py",
        argv=["--res", "1", "--distributed", "--emulate"],
    )
    assert sim.resolution == 32
    assert sim.state["t"] > 0.0
    assert jnp.mean(jnp.abs(sim.state["psi"])) == pytest.approx(162.028, rel=rel_tol)


def test_heating_gas():
    sim = run_example_main("examples/heating_gas/heating_gas.py", argv=["--res", "1"])
    assert sim.resolution == 32
    assert sim.state["t"] > 0.0
    assert jnp.mean(jnp.abs(sim.state["psi"])) == pytest.approx(2573.253, rel=rel_tol)
    assert jnp.mean(jnp.abs(sim.state["vx"])) == pytest.approx(4.6109166, rel=rel_tol)
    assert jnp.mean(jnp.abs(sim.state["vy"])) == pytest.approx(2.9255207, rel=rel_tol)
    assert jnp.mean(jnp.abs(sim.state["vz"])) == pytest.approx(4.1069093, rel=rel_tol)


def test_heating_stars():
    sim = run_example_main(
        "examples/heating_stars/heating_stars.py", argv=["--res", "1"]
    )
    assert sim.resolution == 32
    assert sim.state["t"] > 0.0
    assert jnp.mean(jnp.abs(sim.state["psi"])) == pytest.approx(2574.395, rel=rel_tol)
    assert jnp.mean(jnp.abs(sim.state["vel"][:, 0])) == pytest.approx(
        16.625353, rel=rel_tol
    )
    assert jnp.mean(jnp.abs(sim.state["vel"][:, 1])) == pytest.approx(
        17.345486, rel=rel_tol
    )
    assert jnp.mean(jnp.abs(sim.state["vel"][:, 2])) == pytest.approx(
        18.218296, rel=rel_tol
    )



import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import os


def plot_sim(state, checkpoint_dir, i, params):
    """Plot the simulation state."""

    dynamic_range = params["output"]["plot_dynamic_range"]

    # process distributed data
    if params["physics"]["quantum"]:
        nx = state["psi"].shape[0]
        rho_bar_dm = jnp.mean(jnp.abs(state["psi"]) ** 2)
        rho_proj_dm = jnp.log10(
            jax.experimental.multihost_utils.process_allgather(
                jnp.mean(jnp.abs(state["psi"]) ** 2, axis=2), tiled=True
            )
        ).T
    if params["physics"]["hydro"]:
        nx = state["rho"].shape[0]
        rho_bar_gas = jnp.mean(state["rho"])
        rho_proj_gas = jnp.log10(
            jax.experimental.multihost_utils.process_allgather(
                jnp.mean(state["rho"], axis=2), tiled=True
            )
        ).T

    # create plot on process 0
    if jax.process_index() == 0:
        if params["physics"]["quantum"]:
            plt.clf()

            # DM projection
            nx = state["psi"].shape[0]
            vmin = jnp.log10(rho_bar_dm / dynamic_range)
            vmax = jnp.log10(rho_bar_dm * dynamic_range)

            ax = plt.gca()
            ax.imshow(
                rho_proj_dm,
                cmap="inferno",
                origin="lower",
                vmin=vmin,
                vmax=vmax,
                extent=[0, nx, 0, nx],
            )
            if params["physics"]["particles"]:
                # draw particles
                box_size = params["domain"]["box_size"]
                sx = (state["pos"][:, 0] / box_size) * nx
                sy = (state["pos"][:, 1] / box_size) * nx
                plt.plot(
                    sx, sy, color="cyan", marker=".", linestyle="None", markersize=5
                )
            ax.set_aspect("equal")
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)

            plt.savefig(
                os.path.join(checkpoint_dir, f"dm{i:03d}.png"),
                bbox_inches="tight",
                pad_inches=0,
            )
            plt.close()

        if params["physics"]["hydro"]:
            plt.clf()

            # gas projection
            nx = state["rho"].shape[0]
            vmin = jnp.log10(rho_bar_gas / dynamic_range)
            vmax = jnp.log10(rho_bar_gas * dynamic_range)

            ax = plt.gca()
            ax.imshow(
                rho_proj_gas,
                cmap="viridis",
                origin="lower",
                vmin=vmin,
                vmax=vmax,
                extent=[0, nx, 0, nx],
            )
            if params["physics"]["particles"]:
                # draw particles
                box_size = params["domain"]["box_size"]
                sx = (state["pos"][:, 0] / box_size) * nx
                sy = (state["pos"][:, 1] / box_size) * nx
                plt.plot(
                    sx, sy, color="red", marker=".", linestyle="None", markersize=5
                )
            ax.set_aspect("equal")
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)

            plt.savefig(
                os.path.join(checkpoint_dir, f"gas{i:03d}.png"),
                bbox_inches="tight",
                pad_inches=0,
            )
            plt.close()



import jax.numpy as jnp

# Pure functions for cosmology simulation


def get_physical_time_interval(z_start, z_end, omega_matter, omega_lambda, little_h):
    """Compute the total physical time between two redshifts."""
    #  da/dt = H0 * sqrt(omega_matter / a + omega_lambda * a^2)
    a_start = 1.0 / (1.0 + z_start)
    a_end = 1.0 / (1.0 + z_end)
    H0 = 0.1 * little_h  # Hubble constant in (km/s/kpc)
    n_quad = 10000
    a_lin = jnp.linspace(a_start, a_end, n_quad)
    a_dot = H0 * jnp.sqrt((omega_matter / a_lin) + (omega_lambda * a_lin**2))
    dt_hat = jnp.trapezoid(1.0 / a_dot, a_lin)
    return dt_hat


def get_supercomoving_time_interval(
    z_start, z_end, omega_matter, omega_lambda, little_h
):
    """Compute the total supercomoving time (dt_hat = a^-2 dt) between two redshifts."""
    #  da/dt = H0 * sqrt(omega_matter / a + omega_lambda * a^2)
    #  da/dt_hat = a^2 * da/dt
    #  dt_hat/da = a^-2 / (da/dt)
    a_start = 1.0 / (1.0 + z_start)
    a_end = 1.0 / (1.0 + z_end)
    H0 = 0.1 * little_h  # Hubble constant in (km/s/kpc)
    n_quad = 10000
    a_lin = jnp.linspace(a_start, a_end, n_quad)
    a_dot = H0 * jnp.sqrt((omega_matter / a_lin) + (omega_lambda * a_lin**2))
    dt_hat_da = a_lin**-2 / a_dot
    dt_hat = jnp.trapezoid(dt_hat_da, a_lin)
    return dt_hat


def get_scale_factor(z_start, dt_hat, omega_matter, omega_lambda, little_h):
    """Compute the scale factor corresponding to a given supercomoving time,
    by root finding."""
    a_start = 1.0 / (1.0 + z_start)
    a = a_start
    tolerance = 1e-6
    max_iterations = 100
    lower_bound = 0.0
    upper_bound = 2.0  # Set a reasonable upper bound for the scale factor
    for _ in range(max_iterations):
        dt_hat_guess = get_supercomoving_time_interval(
            z_start, 1.0 / a - 1.0, omega_matter, omega_lambda, little_h
        )
        error = dt_hat_guess - dt_hat
        if jnp.abs(error) < tolerance:
            break
        # Use bisection method
        if error > 0:
            upper_bound = a  # Move upper bound down
        else:
            lower_bound = a  # Move lower bound up
        a = (lower_bound + upper_bound) / 2  # Update scale factor to midpoint
    return a


def get_next_scale_factor(z_start, dt_hat, omega_matter, omega_lambda, little_h):
    """Advance scale factor by dt_hat using RK2 (midpoint) on da/dt_hat = a^2 * da/dt."""
    a_start = 1.0 / (1.0 + z_start)
    H0 = 0.1 * little_h

    def g(a):
        a_dot = H0 * jnp.sqrt((omega_matter / a) + (omega_lambda * a**2))
        return a**2 * a_dot  # da/dt_hat

    k1 = g(a_start)
    k2 = g(a_start + 0.5 * dt_hat * k1)
    a = a_start + dt_hat * k2
    return a



import jax
import jax.numpy as jnp
import orbax.checkpoint as ocp
import os
import json
import time

from .constants import constants
from .quantum import quantum_kick, quantum_drift, quantum_velocity
from .gravity import calculate_gravitational_potential
from .hydro import hydro_fluxes, hydro_accelerate
from .particles import (
    particles_accelerate,
    particles_drift,
    particles_accrete_gas,
    bin_particles,
)
from .cosmology import get_supercomoving_time_interval, get_next_scale_factor
from .utils import (
    set_up_parameters,
    print_parameters,
    print_distributed_info,
    xmeshgrid,
    xmeshgrid_transpose,
    xzeros,
    xones,
)
from .visualization import plot_sim


class Simulation:
    """
    Simulation: The base class for an astrophysics simulation.

    Parameters
    ----------
        params : dict
            The Python dictionary that contains the simulation parameters.
            Params can also be a string path to a checkpoint directory to load a saved simulation.
        sharding : jax.sharding.NamedSharding, optional
            jax sharding used for distributed (multi-GPU) simulations

    """

    def __init__(self, params, sharding=None):
        # allow loading directly from a checkpoint path
        load_from_checkpoint = False
        checkpoint_dir = ""
        if isinstance(params, str):
            load_from_checkpoint = True
            checkpoint_dir = os.path.join(os.getcwd(), params)
            with open(os.path.join(checkpoint_dir, "params.json"), "r") as f:
                params = json.load(f)

        # start from default simulation parameters and update with user params
        self._params = set_up_parameters(params)

        # additional checks
        if self.resolution % 2 != 0:
            raise ValueError("Resolution must be divisible by 2.")

        if self.params["time"]["adaptive"]:
            raise NotImplementedError("Adaptive time stepping is not yet implemented.")

        if self.params["physics"]["cosmology"]:
            if (
                self.params["physics"]["hydro"]
                or self.params["physics"]["particles"]
                or self.params["physics"]["external_potential"]
                or self.params["quantum"]["f_15"] != 0.0
            ):
                raise NotImplementedError(
                    "Cosmological hydro/particles/external_potential/SI is not yet implemented."
                )

        if self.params["physics"]["hydro"] or self.params["physics"]["particles"]:
            if sharding is not None:
                raise NotImplementedError(
                    "hydro/particles sharding is not yet implemented."
                )

        # print info
        if jax.process_index() == 0:
            print("Simulation parameters:")
            print_parameters(self.params)
            if sharding is not None:
                print_distributed_info()

        # jitted functions
        self.xmeshgrid_jit = jax.jit(
            xmeshgrid, in_shardings=None, out_shardings=sharding
        )
        self.xmeshgrid_transpose_jit = jax.jit(
            xmeshgrid_transpose, in_shardings=None, out_shardings=sharding
        )
        self.xzeros_jit = jax.jit(
            xzeros, static_argnums=0, in_shardings=None, out_shardings=sharding
        )
        self.xones_jit = jax.jit(
            xones, static_argnums=0, in_shardings=None, out_shardings=sharding
        )

        # custom functions
        self.custom_kick = None
        self.custom_drift = None
        self.custom_density = None
        self.custom_plot = None

        # simulation state
        self.state = {}
        self.state["t"] = 0.0
        if self.params["physics"]["cosmology"]:
            self.state["redshift"] = 0.0
        if self.params["physics"]["quantum"]:
            self.state["psi"] = self.xzeros_jit(self.resolution) * 1j
        if self.params["physics"]["external_potential"]:
            self.state["V_ext"] = self.xzeros_jit(self.resolution)
        if self.params["physics"]["hydro"]:
            self.state["rho"] = jnp.zeros(
                (self.resolution, self.resolution, self.resolution),
            )
            self.state["vx"] = jnp.zeros(
                (self.resolution, self.resolution, self.resolution),
            )
            self.state["vy"] = jnp.zeros(
                (self.resolution, self.resolution, self.resolution),
            )
            self.state["vz"] = jnp.zeros(
                (self.resolution, self.resolution, self.resolution),
            )
        if self.params["physics"]["particles"]:
            self.state["pos"] = jnp.zeros((self.num_particles, 3))
            self.state["vel"] = jnp.zeros((self.num_particles, 3))
            if self.params["particles"]["accrete_gas"]:
                self.state["mass"] = (
                    jnp.zeros(self.num_particles)
                    + self.params["particles"]["particle_mass"]
                )

        if load_from_checkpoint:
            options = ocp.CheckpointManagerOptions()
            async_checkpoint_manager = ocp.CheckpointManager(
                checkpoint_dir, options=options
            )
            step = async_checkpoint_manager.latest_step()
            self.state = async_checkpoint_manager.restore(
                step, args=ocp.args.StandardRestore(self.state)
            )

    @property
    def resolution(self):
        """
        Return the (linear) resolution of the simulation
        """
        return (
            self.params["domain"]["resolution_base"]
            * self.params["domain"]["resolution_multiplier"]
        )

    @property
    def num_particles(self):
        """
        Return the number of particles in the simulation
        """
        return self.params["particles"]["num_particles"]

    @property
    def box_size(self):
        """
        Return the box size of the simulation (kpc)
        """
        return self.params["domain"]["box_size"]

    @property
    def dx(self):
        """
        Return the cell size size of the simulation (kpc)
        """
        return self.box_size / self.resolution

    @property
    def axion_mass(self):
        """
        Return the axion particle mass in the simulation (M_sun)
        """
        return (
            self.params["quantum"]["m_22"]
            * 1.0e-22
            * constants["electron_volt"]
            / constants["speed_of_light"] ** 2
        )

    @property
    def scattering_length(self):
        """
        Return the axion self-interaction scattering length in the simulation (kpc)
        """
        f_15 = self.params["quantum"]["f_15"]
        if f_15 == 0.0:
            return 0.0
        else:
            f = f_15 * 1.0e24 * constants["electron_volt"]
            sign = 1.0 if f > 0 else -1.0
            hbar = constants["reduced_planck_constant"]
            c = constants["speed_of_light"]
            m = self.axion_mass
            a_s = sign * (hbar * c**3 * m) / (32.0 * jnp.pi * (f**2))
            return a_s

    @property
    def m_per_hbar(self):
        """
        Return the mass per hbar in the simulation (M_sun / hbar)
        """
        return self.axion_mass / constants["reduced_planck_constant"]

    @property
    def sound_speed(self):
        """
        Return the isothermal gas sound speed in the simulation (km/s)
        """
        return self.params["hydro"]["sound_speed"]

    @property
    def params(self):
        """
        Return the parameters of the simulation
        """
        return self._params

    @property
    def grid(self):
        """
        Return the simulation grid
        """
        hx = 0.5 * self.dx
        x_lin = jnp.linspace(hx, self.box_size - hx, self.resolution)
        xx, yy, zz = self.xmeshgrid_jit(x_lin)
        return xx, yy, zz

    @property
    def kgrid(self):
        """
        Return the simulation spectral grid
        """
        nx = self.resolution
        k_lin = (2.0 * jnp.pi / self.box_size) * jnp.arange(-nx / 2, nx / 2)
        kx, ky, kz = self.xmeshgrid_transpose_jit(k_lin)
        kx = jnp.fft.ifftshift(kx)
        ky = jnp.fft.ifftshift(ky)
        kz = jnp.fft.ifftshift(kz)
        return kx, ky, kz

    @property
    def quantum_velocity(self):
        """
        Return the dark matter velocity field from the wavefunction
        """
        return quantum_velocity(self.state["psi"], self.box_size, self.m_per_hbar)

    @property
    def rho_bar(self):
        """
        Return the mean density of the simulation
        """
        return self._calc_rho_bar(self.state)

    def _calc_rho_bar(self, state):
        rho_bar = 0.0
        if self.params["physics"]["quantum"]:
            rho_bar += jnp.mean(jnp.abs(state["psi"]) ** 2)
        if self.params["physics"]["hydro"]:
            rho_bar += jnp.mean(state["rho"])
        if self.params["physics"]["particles"]:
            box_size = self.box_size
            if self.params["particles"]["accrete_gas"]:
                rho_bar += jnp.sum(state["mass"]) / box_size**3
            else:
                m_particle = self.params["particles"]["particle_mass"]
                n_particles = self.num_particles
                rho_bar += m_particle * n_particles / box_size**3
        if self.custom_density is not None:
            rho_bar += jnp.mean(self.custom_density(state))
        return rho_bar

    def _calc_grav_potential(self, state, k_sq):
        G = constants["gravitational_constant"]
        rho_bar = self._calc_rho_bar(state)
        rho_tot = 0.0
        if self.params["physics"]["quantum"]:
            rho_tot += jnp.abs(state["psi"]) ** 2
        if self.params["physics"]["hydro"]:
            rho_tot += state["rho"]
        if self.params["physics"]["particles"]:
            multiple_masses = self.params["particles"]["accrete_gas"]
            if multiple_masses:
                m_particles = state["mass"]
                rho_tot += bin_particles(
                    state["pos"], m_particles, self.dx, self.resolution, multiple_masses
                )
            else:
                m_particles = self.params["particles"]["particle_mass"]
                rho_tot += bin_particles(
                    state["pos"], m_particles, self.dx, self.resolution, multiple_masses
                )
        if self.custom_density is not None:
            rho_tot += self.custom_density(state)
        if self.params["physics"]["cosmology"]:
            scale_factor = 1.0 / (1.0 + state["redshift"])
            rho_bar *= scale_factor
            rho_tot *= scale_factor
        return calculate_gravitational_potential(rho_tot, k_sq, G, rho_bar)

    @property
    def potential(self):
        """
        Return the gravitational potential
        """
        kx, ky, kz = self.kgrid
        k_sq = kx**2 + ky**2 + kz**2
        return self._calc_grav_potential(self.state, k_sq)

    def _evolve(self, state):
        """
        This function evolves the simulation state according to the simulation parameters/physics.

        Parameters
        ----------
        state: jax.pytree
          The current state of the simulation.

        Returns
        -------
        state: jax.pytree
          The evolved state of the simulation.
        """

        # Simulation parameters
        dx = self.dx
        box_size = self.box_size
        num_cells = self.resolution**3
        m_per_hbar = self.m_per_hbar

        safety = self.params["time"]["safety_factor"]
        dt_kin = safety * (m_per_hbar / 6.0) * (dx * dx)
        t_start = self.params["time"]["start"]
        t_end = self.params["time"]["end"]
        t_span = t_end - t_start
        state["t"] = t_start

        use_quantum = self.params["physics"]["quantum"]
        use_gravity = self.params["physics"]["gravity"]
        use_hydro = self.params["physics"]["hydro"]
        use_particles = self.params["physics"]["particles"]
        use_cosmology = self.params["physics"]["cosmology"]
        use_external_potential = self.params["physics"]["external_potential"]
        accrete_gas = self.params["particles"]["accrete_gas"]
        save = self.params["output"]["save"]
        use_custom = self.custom_kick is not None or self.custom_drift is not None
        if use_custom:
            custom_kick = self.custom_kick
            custom_drift = self.custom_drift

        # cosmology
        if use_cosmology:
            z_start = self.params["time"]["start"]
            z_end = self.params["time"]["end"]
            omega_matter = self.params["cosmology"]["omega_matter"]
            omega_lambda = self.params["cosmology"]["omega_lambda"]
            little_h = self.params["cosmology"]["little_h"]
            t_span = get_supercomoving_time_interval(
                z_start, z_end, omega_matter, omega_lambda, little_h
            )
            state["t"] = 0.0
            state["redshift"] = z_start

        # self-interaction
        a_s = self.scattering_length
        c = constants["speed_of_light"]
        m = self.axion_mass
        si_coeff = None
        si_coeff2 = None
        do_self_interaction = a_s != 0.0
        if do_self_interaction:
            si_coeff = (4.0 * jnp.pi) * (a_s / m) / m_per_hbar**2
            si_coeff2 = (32.0 * jnp.pi**2 / 3.0) * (a_s / (m * c)) ** 2 / m_per_hbar**5

        # hydro
        c_sound = self.params["hydro"]["sound_speed"]

        # round up to the nearest multiple of num_checkpoints
        num_checkpoints = self.params["output"]["num_checkpoints"]
        nt = int(round(round(t_span / dt_kin) / num_checkpoints) * num_checkpoints)
        nt_sub = int(round(nt / num_checkpoints))
        dt = t_span / nt

        # distributed arrays (fixed) needed for calculations
        kx, ky, kz = None, None, None
        k_sq = None

        # Fourier space variables
        if use_gravity or use_quantum:
            kx, ky, kz = self.kgrid
            k_sq = kx**2 + ky**2 + kz**2

        # Checkpointer
        if save:
            options = ocp.CheckpointManagerOptions()
            checkpoint_dir = os.path.join(os.getcwd(), self.params["output"]["path"])
            path = os.path.join(os.getcwd(), checkpoint_dir)
            if jax.process_index() == 0:
                path = ocp.test_utils.erase_and_create_empty(checkpoint_dir)
            async_checkpoint_manager = ocp.CheckpointManager(path, options=options)

        carry = (state, kx, ky, kz, k_sq)

        def _kick(state, kx, ky, kz, k_sq, dt):
            # Kick (half-step)
            if use_gravity and use_external_potential:
                V = self._calc_grav_potential(state, k_sq) + state["V_ext"]
            elif use_gravity:
                V = self._calc_grav_potential(state, k_sq)
            elif use_external_potential:
                V = state["V_ext"]

            if use_gravity or use_external_potential:
                if use_quantum:
                    if do_self_interaction:
                        rho = jnp.abs(state["psi"]) ** 2
                        V_prime = V + si_coeff * rho + si_coeff2 * rho**2
                        state["psi"] = quantum_kick(
                            state["psi"], V_prime, m_per_hbar, dt
                        )
                    else:
                        state["psi"] = quantum_kick(state["psi"], V, m_per_hbar, dt)
                if use_hydro:
                    state["vx"], state["vy"], state["vz"] = hydro_accelerate(
                        state["vx"], state["vy"], state["vz"], V, kx, ky, kz, dt
                    )
                if use_particles:
                    state["vel"] = particles_accelerate(
                        state["vel"], state["pos"], V, kx, ky, kz, dx, dt
                    )
                if use_custom:
                    state = custom_kick(state, V, dt)

            return state

        def _drift(state, k_sq, dt):
            # Drift (full-step)
            if use_quantum:
                state["psi"] = quantum_drift(state["psi"], k_sq, m_per_hbar, dt)
            if use_hydro:
                state["rho"], state["vx"], state["vy"], state["vz"] = hydro_fluxes(
                    state["rho"], state["vx"], state["vy"], state["vz"], dt, dx, c_sound
                )
            if use_particles:
                state["pos"] = particles_drift(state["pos"], state["vel"], dt, box_size)
            if use_custom:
                state = custom_drift(state, k_sq, dt)
            if use_hydro and accrete_gas:
                G = constants["gravitational_constant"]
                state["mass"], state["rho"] = particles_accrete_gas(
                    state["mass"], state["rho"], state["pos"], G, c_sound, dx, dt
                )

            return state

        def _update(_, carry):
            # Update the simulation state by one timestep
            # according to a 2nd-order `kick-drift-kick` scheme
            state, kx, ky, kz, k_sq = carry
            state = _kick(state, kx, ky, kz, k_sq, 0.5 * dt)
            state = _drift(state, k_sq, dt)
            # update time & redshift
            state["t"] += dt
            if use_cosmology:
                scale_factor = get_next_scale_factor(
                    state["redshift"], dt, omega_matter, omega_lambda, little_h
                )
                state["redshift"] = 1.0 / scale_factor - 1.0
            state = _kick(state, kx, ky, kz, k_sq, 0.5 * dt)

            return state, kx, ky, kz, k_sq

        # save initial state
        if jax.process_index() == 0:
            print(f"Starting simulation (res={self.resolution}, nt={nt}) ...")
        if save:
            with open(os.path.join(checkpoint_dir, "params.json"), "w") as f:
                json.dump(self.params, f, indent=2)
            async_checkpoint_manager.save(0, args=ocp.args.StandardSave(state))
            plot_sim(state, checkpoint_dir, 0, self.params)
            if self.custom_plot is not None:
                self.custom_plot(state, checkpoint_dir, 0, self.params)
            async_checkpoint_manager.wait_until_finished()

        # Simulation Main Loop
        t_start_timer = time.time()
        if save:
            for i in range(1, num_checkpoints + 1):
                carry = jax.lax.fori_loop(0, nt_sub, _update, init_val=carry)
                state, _, _, _, _ = carry
                jax.block_until_ready(state)
                # save state
                async_checkpoint_manager.save(i, args=ocp.args.StandardSave(state))
                percent = int(100 * i / num_checkpoints)
                elapsed = time.time() - t_start_timer
                est_total = elapsed / i * num_checkpoints
                est_remaining = est_total - elapsed
                mcups = (num_cells * (i * nt_sub)) / (elapsed * 1.0e6)
                if jax.process_index() == 0:
                    print(
                        f"{percent:.1f}%: mcups={mcups:.1f}, estimated time left (s): {est_remaining:.1f}"
                    )
                plot_sim(state, checkpoint_dir, i, self.params)
                if self.custom_plot is not None:
                    self.custom_plot(state, checkpoint_dir, i, self.params)
                async_checkpoint_manager.wait_until_finished()
        else:
            carry = jax.lax.fori_loop(0, nt, _update, init_val=carry)
            state, _, _, _, _ = carry
        jax.block_until_ready(state)
        if jax.process_index() == 0:
            print("Simulation Run Time (s): ", time.time() - t_start_timer)

        return state

    def run(self):
        """
        Run the simulation
        """
        self.state = self._evolve(self.state)
        jax.block_until_ready(self.state)



import jax.numpy as jnp
import jaxion
from jaxion.quantum import quantum_velocity
from jaxion.analysis import radial_power_spectrum
import pytest


def test_quantum_velocity_and_radial_power_spectrum():
    sim = jaxion.Simulation({})
    box_size = sim.box_size
    m_per_hbar = sim.m_per_hbar
    xx, yy, _ = sim.grid
    kx, ky, kz = sim.kgrid

    psi = (
        jnp.cos(2.0 * jnp.pi * xx / box_size) ** 2
        + jnp.cos(2.0 * jnp.pi * yy / box_size) * 1j
    )

    vx, _, _ = quantum_velocity(psi, box_size, m_per_hbar)

    Pf, _, _ = radial_power_spectrum(vx, kx, ky, kz, box_size)

    assert jnp.max(Pf) == pytest.approx(8244.606, rel=1e-4)



import jax.numpy as jnp
import jaxdecomp as jd


def radial_power_spectrum(data_cube, kx, ky, kz, box_size):
    """
    Computes the radially averaged power spectral density of a 3D datacube.

    Parameters
    ----------
        data_cube : jnp.ndarray
            3D data cube, must be cubic
        kx, ky, kz: jnp.ndarray
            wavenumber grids in each dimension
        box_size: float
            physical size of box

    Returns
    -------
        Pf: jnp.ndarray
            radial power spectrum
        k: jnp.ndarray
            wavenumbers
        total_power: float
            total power
    """
    dim = data_cube.ndim
    nx = data_cube.shape[0]
    dx = box_size / nx

    # Compute power spectrum
    data_cube_hat = jd.fft.pfft3d(data_cube)
    total_power = 0.5 * jnp.sum(jnp.abs(data_cube_hat) ** 2) / nx**dim * dx**dim
    phi_k = 0.5 * jnp.abs(data_cube_hat) ** 2 / nx**dim * dx**dim
    half_size = nx // 2 + 1

    # Compute radially-averaged power spectrum
    # if dim == 2:
    #    k_r = jnp.sqrt(kx**2 + ky**2)
    k_r = jnp.sqrt(kx**2 + ky**2 + kz**2)

    Pf, _ = jnp.histogram(
        k_r, range=(-0.5, half_size - 0.5), bins=half_size, weights=phi_k
    )
    norm, _ = jnp.histogram(k_r, range=(-0.5, half_size - 0.5), bins=half_size)
    Pf /= norm + (norm == 0)

    k = 2.0 * jnp.pi * jnp.arange(half_size) / box_size
    dk = 2.0 * jnp.pi / box_size

    Pf /= dk**dim

    # Add geometrical factor
    # if dim == 2:
    #     Pf = Pf * 2.0 * jnp.pi * k
    Pf *= 4.0 * jnp.pi * k**2

    return Pf, k, total_power



from jaxion.cosmology import (
    get_physical_time_interval,
    get_supercomoving_time_interval,
    get_scale_factor,
    get_next_scale_factor,
)
import pytest


def test_cosmology_functions():
    z_start = 127.0
    z_end = 0.0
    omega_matter = 0.3
    omega_lambda = 0.7
    little_h = 0.7
    dt_hat = 10.0

    assert get_physical_time_interval(
        z_start, z_end, omega_matter, omega_lambda, little_h
    ) == pytest.approx(13.76084)
    assert get_supercomoving_time_interval(
        z_start, z_end, omega_matter, omega_lambda, little_h
    ) == pytest.approx(530.44415)
    assert get_scale_factor(
        z_start, dt_hat, omega_matter, omega_lambda, little_h
    ) == pytest.approx(0.008084139320999384)
    assert get_next_scale_factor(
        z_start, dt_hat, omega_matter, omega_lambda, little_h
    ) == pytest.approx(0.00808401)



# Copyright (c) 2025 The Jaxion Team.

"""
Physical constants in units of:
[L] = kpc,
[V] = km/s,
[M] = Msun

note: other units are derived from these base units, e.g., [T] = [L]/[V] = kpc / (km/s) ~= 0.978 Gyr
"""

constants = {
    "gravitational_constant": 4.30241002e-6,  # G / (kpc * (km/s)^2 / Msun)
    "reduced_planck_constant": 1.71818134e-87,  # hbar / ((km/s) * kpc * mass of sun)
    "electron_volt": 8.05478173e-56,  # eV / (mass of sun * (km/s)^2)
    "speed_of_light": 299792.458,  # c / (km/s)
}



import jax.numpy as jnp
import jaxdecomp as jd

# Pure functions for gravity calculations


def calculate_gravitational_potential(rho, k_sq, G, rho_bar):
    V_hat = -jd.fft.pfft3d(4.0 * jnp.pi * G * (rho - rho_bar)) / (k_sq + (k_sq == 0))
    V = jnp.real(jd.fft.pifft3d(V_hat))
    return V



import sys
import os
import importlib.util
from pathlib import Path
import importlib.resources
import json
from importlib.metadata import version
import jax
import jax.numpy as jnp


def print_parameters(params):
    print(json.dumps(params, indent=2))


def print_distributed_info():
    for env_var in [
        "SLURM_JOB_ID",
        "SLURM_NTASKS",
        "SLURM_NODELIST",
        "SLURM_STEP_NODELIST",
        "SLURM_STEP_GPUS",
        "SLURM_GPUS",
    ]:
        print(f"{env_var}: {os.getenv(env_var, '')}")
    print("Total number of processes: ", jax.process_count())
    print("Total number of devices: ", jax.device_count())
    print("List of devices: ", jax.devices())
    print("Number of devices on this process: ", jax.local_device_count())


def set_up_parameters(user_overwrites):
    # first load the default params
    params_path = importlib.resources.files("jaxion") / "defaults.json"
    with params_path.open("r", encoding="utf-8") as f:
        params = json.load(f)

    # go down to lowest level in dict and eliminate meta-data
    params = _eliminate_metadata(params)

    # update default values with user-supplied overwrites
    params = _update_dicts(params, user_overwrites)

    # detect jaxion version
    params["version"] = version("jaxion")

    return params


def _eliminate_metadata(params):
    for key, value in params.items():
        if isinstance(value, dict):
            if "default" not in value:
                _eliminate_metadata(value)
            else:
                params[key] = value["default"]

    return params


def _update_dicts(orig_dict, new_dict):
    for key, value in new_dict.items():
        if (
            key in orig_dict
            and isinstance(orig_dict[key], dict)
            and isinstance(value, dict)
        ):
            _update_dicts(orig_dict[key], value)
        else:
            if key in orig_dict:
                if not isinstance(value, dict):
                    orig_dict[key] = value
                else:
                    raise ValueError(
                        f"Value: {value} for parameter key: {key} must be a value, not dict"
                    )
            else:
                raise KeyError(f"Unknown parameter key: {key}")
    return orig_dict


def run_example_main(example_path, argv=None):
    """
    Utility to run an example script's main() as if from its directory, with optional argv patching.
    Args:
        example_path (str or Path): Path to the example script (e.g., 'examples/foo/foo.py')
        argv (list, optional): List of arguments to patch sys.argv with. If None, uses [script_name].
    """
    example_path = Path(example_path)
    example_dir = example_path.parent
    script_name = example_path.name
    old_cwd = os.getcwd()
    old_argv = sys.argv.copy()
    try:
        spec = importlib.util.spec_from_file_location(
            script_name.rstrip(".py"), example_path
        )
        os.chdir(example_dir)
        sys.argv = [script_name] + (argv if argv is not None else [])
        module = importlib.util.module_from_spec(spec)
        sys.modules[script_name.rstrip(".py")] = module
        spec.loader.exec_module(module)
        if hasattr(module, "main"):
            result = module.main()
        else:
            raise AttributeError(f"No main() in {example_path}")
    finally:
        os.chdir(old_cwd)
        sys.argv = old_argv
    return result


# Make a distributed meshgrid function
def xmeshgrid(x_lin):
    xx, yy, zz = jnp.meshgrid(x_lin, x_lin, x_lin, indexing="ij")
    return xx, yy, zz


# NOTE: jaxdecomp (jd) has pfft3d transpose the axis (X, Y, Z) --> (Y, Z, X), and pifft3d undo it
# so the fourier space variables (e.g. kx, ky, kz) all need to be transposed


def xmeshgrid_transpose(x_lin):
    xx, yy, zz = jnp.meshgrid(x_lin, x_lin, x_lin, indexing="ij")
    xx = jnp.transpose(xx, (1, 2, 0))
    yy = jnp.transpose(yy, (1, 2, 0))
    zz = jnp.transpose(zz, (1, 2, 0))
    return xx, yy, zz


def xzeros(nx):
    return jnp.zeros((nx, nx, nx))


def xones(nx):
    return jnp.ones((nx, nx, nx))
